"""
    picasso/lib
    ~~~~~~~~~~~~~~~~~~~~

    Handy functions and classes

    :author: Joerg Schnitzbauer, 2016
    :copyright: Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry
"""


import numba as _numba
import numpy as _np
from numpy.lib.recfunctions import append_fields as _append_fields
from numpy.lib.recfunctions import drop_fields as _drop_fields
import collections as _collections
import glob as _glob
import os.path as _ospath
from picasso import io as _io
from PyQt5 import QtGui, QtCore, QtWidgets
from lmfit import Model as _Model


# A global variable where we store all open progress and status dialogs.
# In case of an exception, we close them all,
# so that the GUI remains responsive.
_dialogs = []


class ProgressDialog(QtWidgets.QProgressDialog):
    def __init__(self, description, minimum, maximum, parent):
        super().__init__(
            description,
            None,
            minimum,
            maximum,
            parent,
            QtCore.Qt.CustomizeWindowHint,
        )
        self.initalized = None

    def init(self):
        _dialogs.append(self)
        self.setMinimumDuration(500)
        self.setModal(True)
        self.app = QtCore.QCoreApplication.instance()
        self.initalized = True

    def set_value(self, value):
        if not self.initalized:
            self.init()
        self.setValue(value)
        self.app.processEvents()

    def closeEvent(self, event):
        _dialogs.remove(self)


class StatusDialog(QtWidgets.QDialog):
    def __init__(self, description, parent):
        super(StatusDialog, self).__init__(parent, QtCore.Qt.CustomizeWindowHint)
        _dialogs.append(self)
        vbox = QtWidgets.QVBoxLayout(self)
        label = QtWidgets.QLabel(description)
        vbox.addWidget(label)
        self.show()
        QtCore.QCoreApplication.instance().processEvents()

    def closeEvent(self, event):
        _dialogs.remove(self)


class AutoDict(_collections.defaultdict):
    """
    A defaultdict whose auto-generated values are defaultdicts itself.
    This allows for auto-generating nested values, e.g.
    a = AutoDict()
    a['foo']['bar']['carrot'] = 42
    """

    def __init__(self, *args, **kwargs):
        super().__init__(AutoDict, *args, **kwargs)


def cancel_dialogs():
    dialogs = [_ for _ in _dialogs]
    for dialog in dialogs:
        if isinstance(dialog, ProgressDialog):
            dialog.cancel()
        else:
            dialog.close()
    QtCore.QCoreApplication.instance().processEvents()  # just in case...


def cumulative_exponential(x, a, t, c):
    return a * (1 - _np.exp(-(x / t))) + c


CumulativeExponentialModel = _Model(cumulative_exponential)


def calculate_optimal_bins(data, max_n_bins=None):
    iqr = _np.subtract(*_np.percentile(data, [75, 25]))
    bin_size = 2 * iqr * len(data) ** (-1 / 3)
    if data.dtype.kind in ("u", "i") and bin_size < 1:
        bin_size = 1
    bin_min = data.min() - bin_size / 2
    n_bins = _np.ceil((data.max() - bin_min) / bin_size)
    try:
        n_bins = int(n_bins)
    except ValueError:
        return None
    if max_n_bins and n_bins > max_n_bins:
        n_bins = max_n_bins
    return _np.linspace(bin_min, data.max(), n_bins)


def append_to_rec(rec_array, data, name):
    if hasattr(rec_array, name):
        rec_array = remove_from_rec(rec_array, name)
    return _append_fields(
        rec_array,
        name,
        data,
        dtypes=data.dtype,
        usemask=False,
        asrecarray=True,
    )
    return rec_array

def ensure_sanity(locs, info):
    # no inf or nan:
    locs = locs[
        _np.all(
            _np.array([_np.isfinite(locs[_]) for _ in locs.dtype.names]),
            axis=0,
        )
    ]
    # other sanity checks:
    locs = locs[locs.x > 0]
    locs = locs[locs.y > 0]
    locs = locs[locs.x < info[0]["Width"]]
    locs = locs[locs.y < info[0]["Height"]]
    locs = locs[locs.lpx > 0]
    locs = locs[locs.lpy > 0]
    return locs


def is_loc_at(x, y, locs, r):
    dx = locs.x - x
    dy = locs.y - y
    r2 = r**2
    return dx**2 + dy**2 < r2


def locs_at(x, y, locs, r):
    is_picked = is_loc_at(x, y, locs, r)
    return locs[is_picked]


@_numba.jit(nopython=True)
def check_if_in_rectangle(x, y, X, Y):
    """
    Checks if locs with coordinates (x, y) are in rectangle with corners (X, Y)
    by counting the number of rectangle sides which are hit by a ray
    originating from each loc to the right. If the number of hit rectangle
    sides is odd, then the loc is in the rectangle
    """
    n_locs = len(x)
    ray_hits_rectangle_side = _np.zeros((n_locs, 4))
    for i in range(4):
        # get two y coordinates of corner points forming one rectangle side
        y_corner_1 = Y[i]
        # take the first if we're at the last side:
        y_corner_2 = Y[0] if i == 3 else Y[i + 1]
        y_corners_min = min(y_corner_1, y_corner_2)
        y_corners_max = max(y_corner_1, y_corner_2)
        for j in range(n_locs):
            y_loc = y[j]
            # only if loc is on level of rectangle side, its ray can hit:
            if y_corners_min <= y_loc <= y_corners_max:
                x_corner_1 = X[i]
                # take the first if we're at the last side:
                x_corner_2 = X[0] if i == 3 else X[i + 1]
                # calculate intersection point of ray and side:
                m_inv = (x_corner_2 - x_corner_1) / (y_corner_2 - y_corner_1)
                x_intersect = m_inv * (y_loc - y_corner_1) + x_corner_1
                x_loc = x[j]
                if x_intersect >= x_loc:
                    # ray hits rectangle side on the right side
                    ray_hits_rectangle_side[j, i] = 1
    n_sides_hit = _np.sum(ray_hits_rectangle_side, axis=1)
    is_in_rectangle = n_sides_hit % 2 == 1
    return is_in_rectangle


def locs_in_rectangle(locs, X, Y):
    is_in_rectangle = check_if_in_rectangle(locs.x, locs.y, _np.array(X), _np.array(Y))
    return locs[is_in_rectangle]


def minimize_shifts(shifts_x, shifts_y, shifts_z=None):
    n_channels = shifts_x.shape[0]
    n_pairs = int(n_channels * (n_channels - 1) / 2)
    n_dims = 2 if shifts_z is None else 3
    rij = _np.zeros((n_pairs, n_dims))
    A = _np.zeros((n_pairs, n_channels - 1))
    flag = 0
    for i in range(n_channels - 1):
        for j in range(i + 1, n_channels):
            rij[flag, 0] = shifts_y[i, j]
            rij[flag, 1] = shifts_x[i, j]
            if n_dims == 3:
                rij[flag, 2] = shifts_z[i, j]
            A[flag, i:j] = 1
            flag += 1
    Dj = _np.dot(_np.linalg.pinv(A), rij)
    shift_y = _np.insert(_np.cumsum(Dj[:, 0]), 0, 0)
    shift_x = _np.insert(_np.cumsum(Dj[:, 1]), 0, 0)
    if n_dims == 2:
        return shift_y, shift_x
    else:
        shift_z = _np.insert(_np.cumsum(Dj[:, 2]), 0, 0)
        return shift_y, shift_x, shift_z


def n_futures_done(futures):
    return sum([_.done() for _ in futures])


def remove_from_rec(rec_array, name):
    return _drop_fields(rec_array, name, usemask=False, asrecarray=True)


def locs_glob_map(func, pattern, args=[], kwargs={}, extension=""):
    """
    Maps a function to localization files, specified by a unix style path
    pattern.
    The function must take two arguments: locs and info. It may take additional
    args and kwargs which are supplied to this map function.
    A new locs file will be saved if an extension is provided. In that case the
    mapped function must return new locs and a new info dict.
    """
    paths = _glob.glob(pattern)
    for path in paths:
        locs, info = _io.load_locs(path)
        result = func(locs, info, path, *args, **kwargs)
        if extension:
            base, ext = _ospath.splitext(path)
            out_path = base + "_" + extension + ".hdf5"
            locs, info = result
            _io.save_locs(out_path, locs, info)



"""
    picasso.design
    ~~~~~~~~~~~~~~~~

    Design rectangular rothemund origami (RRO)

    :author: Maximilian Thomas Strauss, 2016-2018
    :copyright: Copyright (c) 2016-2018 Jungmann Lab, MPI of Biochemistry
"""
import csv
from . import io as _io


def saveInfo(filename, info):
    _io.save_info(filename, [info], default_flow_style=True)


def convertPlateIndex(plate, platename):
    # convert from canvas index [CANVAS_INDEX, OLIGONAME, SEQUENCE]
    # format for ordering [PLATE NAME, PLATE POSITION, OLIGONAME, SEQUENCE]

    platerow = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
    ]
    platecol = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    structurerow = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
    ]
    structurecol = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

    newplate = [["PLATE NAME", "PLATE POSITION", "OLIGO NAME", "SEQUENCE"]]
    for row in range(0, len(platerow)):
        for col in range(0, len(platecol)):
            if row < 8:
                platenameindex = platename + "_1"
            else:
                platenameindex = platename + "_2"
            structureindex = structurerow[row] + str(structurecol[col])
            oligoname = " "
            sequence = " "
            for i in range(0, len(plate)):
                if plate[i][0] == structureindex:
                    oligoname = plate[i][1]
                    sequence = plate[i][2]
            newplate.append(
                [
                    platenameindex,
                    platerow[row] + str(platecol[col]),
                    oligoname,
                    sequence,
                ]
            )

    return newplate


def convertPlateIndexColor(plate, platename):
    # convert from canvas index [CANVAS_INDEX, OLIGONAME, SEQUENCE]
    # format for ordering [PLATE NAME, PLATE POSITION, OLIGONAME, SEQUENCE]

    platerow = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
    ]
    platecol = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    structurerow = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
    ]
    structurecol = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

    newplate = [["PLATE NAME", "PLATE POSITION", "OLIGO NAME", "SEQUENCE", "COLOR"]]
    for row in range(0, len(platerow)):
        for col in range(0, len(platecol)):
            if row < 8:
                platenameindex = platename + "_1"
            else:
                platenameindex = platename + "_2"
            structureindex = structurerow[row] + str(structurecol[col])
            oligoname = " "
            sequence = " "
            for i in range(0, len(plate)):
                if plate[i][0] == structureindex:
                    oligoname = plate[i][1]
                    sequence = plate[i][2]
                    color = plate[i][3]
            newplate.append(
                [
                    platenameindex,
                    platerow[row] + str(platecol[col]),
                    oligoname,
                    sequence,
                    color,
                ]
            )

    return newplate


def readPlate(filename):
    File = open(filename)
    Reader = csv.reader(File)
    data = list(Reader)
    return data


def savePlate(filename, data):
    with open(filename, "w", newline="") as csvfile:
        Writer = csv.writer(
            csvfile, delimiter=",", quotechar="|", quoting=csv.QUOTE_MINIMAL
        )
        for j in range(0, len(data)):
            exportdata = data[j]
            for i in range(0, len(exportdata)):
                Writer.writerow(exportdata[i])



import streamlit as st
from helper import _db_filename
from sqlalchemy import create_engine
import pandas as pd
import os
import numpy as np
from picasso import io
from picasso import render
import matplotlib.pyplot as plt


@st.cache
def load_file(path: str):
    """Loads localization from files. Cached version.

    Args:
        path (str): Path to file.
    """
    locs, info = io.load_locs(path)
    return locs, info


@st.cache
def picasso_render(locs: np.ndarray, viewport: tuple, oversampling: float):
    """Helper function to render a viewport. Cached.

    Args:
        locs (np.ndarray): Record array with localization data.
        viewport (tuple): Viewport as tuple.
        oversampling (int): Oversampling.
    """
    len_x, image = render.render(
        locs,
        viewport=viewport,
        oversampling=oversampling,
        blur_method="smooth",
    )

    return image


def preview():
    """
    Streamlit page to preview a file.
    """
    st.write("# Preview")

    st.write(
        "Select a movie from the database. All hdf files with the same base path as the movie will be selectable."
    )
    engine = create_engine("sqlite:///" + _db_filename(), echo=False)

    try:
        df = pd.read_sql_table("files", con=engine)

        file = st.selectbox("Select file", df["filename"].unique())
        # Find files in familiy
        base = os.path.split(file)[1].split(".")[0]
        folder = os.path.dirname(file)

        if os.path.isdir(folder):

            files = os.listdir(folder)
            files = [f for f in files if f.startswith(base) and f.endswith(".hdf5")]

            hdf_file = st.selectbox("Select hdf file", [None] + files)

            if hdf_file is not None:
                st.write("## File preview")

                st.info(
                    "Performance Warning: This preview will render the full image, so it might be slow for large oversmapling."
                )

                with st.spinner("Loading file"):
                    hdf_file_ = os.path.join(folder, hdf_file)
                    if os.path.isfile(hdf_file_):
                        locs, info = load_file(hdf_file_)

                        x_min = np.min(locs.x)
                        x_max = np.max(locs.x)
                        y_min = np.min(locs.y)
                        y_max = np.max(locs.y)

                        viewport = (y_min, x_min), (y_max, x_max)

                        c1, c2, c3 = st.columns(3)

                        oversampling = c1.number_input(
                            "Oversampling", value=5.0, min_value=1, max_value=40
                        )

                        image = picasso_render(locs, viewport, oversampling)

                        vmin = c2.number_input(
                            "Min density", value=np.min(image.flatten())
                        )
                        vmax = c3.number_input(
                            "Max density", value=np.max(image.flatten())
                        )

                        # plt.imshow(image, cmap='hot', vmax=10)
                        fig, ax = plt.subplots()
                        st.write(f"Image with dimensions {image.shape}")
                        im = ax.imshow(image, cmap="hot", vmin=vmin, vmax=vmax)
                        # Hide grid lines
                        ax.grid(False)

                        # Hide axes ticks
                        ax.set_xticks([])
                        ax.set_yticks([])
                        st.pyplot(fig)

                        st.write(df[df["filename"] == file].iloc[0].to_dict())

                    else:
                        st.warning(f"File {hdf_file} not found.")
        else:
            st.error(
                f"Couldn't find folder {folder}. Please check if folder was deleted or moved."
            )

    except ValueError:
        st.write("Database empty. Process files first.")



"""
    picasso/imageprocess
    ~~~~~~~~~~~~~~~~~~~~

    Image processing functions

    :author: Joerg Schnitzbauer, 2016
    :copyright: Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry
"""
import matplotlib.pyplot as _plt
import numpy as _np
from numpy import fft as _fft
import lmfit as _lmfit
from tqdm import tqdm as _tqdm
from . import lib as _lib


_plt.style.use("ggplot")


def xcorr(imageA, imageB):
    FimageA = _fft.fft2(imageA)
    CFimageB = _np.conj(_fft.fft2(imageB))
    return _fft.fftshift(_np.real(_fft.ifft2((FimageA * CFimageB)))) / _np.sqrt(
        imageA.size
    )


def get_image_shift(imageA, imageB, box, roi=None, display=False):
    """Computes the shift from imageA to imageB"""
    if (_np.sum(imageA) == 0) or (_np.sum(imageB) == 0):
        return 0, 0
    # Compute image correlation
    XCorr = xcorr(imageA, imageB)
    # Cut out center roi
    Y, X = imageA.shape
    if roi is not None:
        Y_ = int((Y - roi) / 2)
        X_ = int((X - roi) / 2)
        if Y_ > 0:
            XCorr = XCorr[Y_:-Y_, :]
        else:
            Y_ = 0
        if X_ > 0:
            XCorr = XCorr[:, X_:-X_]
        else:
            X_ = 0
    else:
        Y_ = X_ = 0
    # A quarter of the fit ROI
    fit_X = int(box / 2)
    # A coordinate grid for the fitting ROI
    y, x = _np.mgrid[-fit_X : fit_X + 1, -fit_X : fit_X + 1]
    # Find the brightest pixel and cut out the fit ROI
    y_max_, x_max_ = _np.unravel_index(XCorr.argmax(), XCorr.shape)
    FitROI = XCorr[
        y_max_ - fit_X : y_max_ + fit_X + 1,
        x_max_ - fit_X : x_max_ + fit_X + 1,
    ]

    dimensions = FitROI.shape

    if 0 in dimensions or dimensions[0] != dimensions[1]:
        xc, yc = 0, 0
    else:
        # The fit model
        def flat_2d_gaussian(a, xc, yc, s, b):
            A = a * _np.exp(-0.5 * ((x - xc) ** 2 + (y - yc) ** 2) / s**2) + b
            return A.flatten()

        gaussian2d = _lmfit.Model(
            flat_2d_gaussian, name="2D Gaussian", independent_vars=[]
        )

        # Set up initial parameters and fit
        params = _lmfit.Parameters()
        params.add("a", value=FitROI.max(), vary=True, min=0)
        params.add("xc", value=0, vary=True)
        params.add("yc", value=0, vary=True)
        params.add("s", value=1, vary=True, min=0)
        params.add("b", value=FitROI.min(), vary=True, min=0)
        results = gaussian2d.fit(FitROI.flatten(), params)

        # Get maximum coordinates and add offsets
        xc = results.best_values["xc"]
        yc = results.best_values["yc"]
        xc += X_ + x_max_
        yc += Y_ + y_max_

        if display:
            _plt.figure(figsize=(17, 10))
            _plt.subplot(1, 3, 1)
            _plt.imshow(imageA, interpolation="none")
            _plt.subplot(1, 3, 2)
            _plt.imshow(imageB, interpolation="none")
            _plt.subplot(1, 3, 3)
            _plt.imshow(XCorr, interpolation="none")
            _plt.plot(xc, yc, "x")
            _plt.show()

        xc -= _np.floor(X / 2)
        yc -= _np.floor(Y / 2)

    return -yc, -xc


def rcc(segments, max_shift=None, callback=None):
    n_segments = len(segments)
    shifts_x = _np.zeros((n_segments, n_segments))
    shifts_y = _np.zeros((n_segments, n_segments))
    n_pairs = int(n_segments * (n_segments - 1) / 2)
    flag = 0
    with _tqdm(
        total=n_pairs, desc="Correlating image pairs", unit="pairs"
    ) as progress_bar:
        if callback is not None:
            callback(0)
        for i in range(n_segments - 1):
            for j in range(i + 1, n_segments):
                progress_bar.update()
                shifts_y[i, j], shifts_x[i, j] = get_image_shift(
                    segments[i], segments[j], 5, max_shift
                )
                flag += 1
                if callback is not None:
                    callback(flag)
    return _lib.minimize_shifts(shifts_x, shifts_y)
